// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.inf.usi.ch/carzaniga/siena/
//
//  Authors: Antonio Carzaniga
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2003-2004 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to one of the authors.
//
//
// $Id: bv_table.cc,v 1.6 2010-03-11 15:16:39 carzanig Exp $
//
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef DEBUG_OUTPUT
#include <iostream>
#endif

#include <siena/forwarding.h>
#include <siena/allocator.h>
#include <siena/bvtable.h>

#include "b_predicate.h"
#include "bset_encoding.h"
#include "bset.h"
#include "mmhwrapper.h"
#include "bitvector.h"
#include "bitcounter.h"

#ifdef WITH_STATISTICS
#include "statistics.h"
#endif

namespace siena {

#ifdef DEBUG_OUTPUT
#define DBG(x) {std::cout << x ;}
#else
#define DBG(x) 
#endif

#ifdef DEBUG_OUTPUT
    std::ostream & operator << (std::ostream & os, const bitvector & b) {
	for(unsigned int i = 0; i < b.get_size(); ++i) 
	    os << (b[i] ? '1' : '0');
	return os;
    }
#endif

/** @brief a predicate in a BVTable.
 *
 *  This is an extremely simple data structure.  The predicate is
 *  represented as a linked list of filters.  Then, each predicate
 *  refers directly to its interface id, and is itself a link in a
 *  linked list of predicates, which is what makes up the BVTable.
 *
 *  The consolidation of the data structure builds up a matrix of
 *  bitvectors masks.  Each line i in the matrix corresponds to the
 *  i-th 8-bit block in the encoded message.  Specifically, for each
 *  8-bit value v, the i-th row contains a bitvector representing the
 *  set of filters in this predicate that would _not_ match if value v
 *  was found at position i.  
 **/
struct BVPredicate {
    BVPredicate * next;
    ifid_t id;
    bitvector * masks[bset_t::B8Size][256];

    BVPredicate(BVPredicate *n, ifid_t i) : next(n), id(i) {}

    bool matches(const bset_t & b) const throw();
};

static inline bool covers(unsigned char x, unsigned char y) {
    return ((x & y) == y);
} 

BVTable::BVTable() throw() : plist(0) {}
BVTable::~BVTable() throw() {}

void BVTable::ifconfig(if_t id, const predicate & p) throw(bad_constraint) {
#ifdef WITH_STATISTICS
    sff_timer_stack.push(ifconfig_timer);
#endif
    FTAllocator tmp_mem;
    BFilter * flist = encode_predicate<BFilter>(p, tmp_mem);
    if (!flist) {
#ifdef WITH_STATISTICS
	sff_timer_stack.pop();
#endif
	return;
    }
    plist = new (memory)BVPredicate(plist, id);

    unsigned int flist_size = 0;
    for(BFilter * curs = flist; (curs); curs = curs->next)
	++flist_size;

    for(unsigned int pos = 0; pos < bset_t::B8Size; ++pos) {
	for(unsigned int v = 0; v < 256; ++v) {
	    FABitvector * b = new (memory)FABitvector(memory, 
						      flist_size, true);
	    unsigned int fpos = 0;
	    for(BFilter * curs = flist; (curs); curs = curs->next) {
		if (!covers(v, curs->b.bv8(pos))) 
		    b->clear(fpos);
		++fpos;
	    }
	    plist->masks[pos][v] = b; // TO BE CHANGED WITH A PATRICIA TRIE DB 
	}
    }
#ifdef WITH_STATISTICS
    sff_timer_stack.pop();
#endif
}

// the matching function for a predicate represented as an XDD is
// trivial: starting at the root of the XDD, we simply look at the
// current block in the input set 
//
bool BVPredicate::matches(const bset_t & b) const throw() {
    bitvector res(*masks[0][b.bv8(0)]);
    if (res.get_count() == 0) {
	DBG("#" << std::endl);
	return false;
    } 

    unsigned pos = 1; 
    do {
	DBG('[' << pos << "] " << res << " / " 
	    << (*masks[pos][b.bv8(pos)]) << std::endl);
	res &= *masks[pos][b.bv8(pos)];
	if (res.get_count() == 0) {
	    DBG("#" << std::endl);
	    return false;
	}
	++pos;
    } while (pos < bset_t::B8Size);
    DBG("!" << std::endl);
    return true;
}

void BVTable::consolidate() throw() { }

void BVTable::match(const message & m, MatchHandler & h) const throw() {
    DBG("match: " << std::endl);
#ifdef WITH_STATISTICS
    sff_timer_stack.push(bloom_encoding_timer);
#endif
    bset_t b;
    encode_message(b, &m);
    
    DBG("> " <<  b << std::endl);
#ifdef WITH_STATISTICS
    sff_timer_stack.pop();
    sff_timer_stack.push(match_timer);
#endif
    for (BVPredicate * p = plist; (p); p = p->next) {
	if (p->matches(b)) {
#ifdef WITH_STATISTICS
	    sff_timer_stack.push(forward_timer);
#endif
	    bool output_result = h.output(p->id);
#ifdef WITH_STATISTICS
	    sff_timer_stack.pop();
#endif
	    if (output_result) {
#ifdef WITH_STATISTICS
		sff_timer_stack.pop();
#endif
		return;
	    }
	}
    }
#ifdef WITH_STATISTICS
    sff_timer_stack.pop();
#endif
}

void BVTable::match(const message & m, MatchMessageHandler & p) const throw() {
    MMHWrapper w(p, m);
    match(m, w);
}

void BVTable::clear() throw() {
    memory.clear();
    plist = 0;
}
void BVTable::clear_recycle() throw() {
    memory.recycle();
    plist = 0;
}

size_t BVTable::allocated_bytesize() const throw() {
    return memory.allocated_size();
}

size_t BVTable::bytesize() const throw() {
    return memory.size();
}

} // end namespace siena

