// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.inf.usi.ch/carzaniga/siena/
//
//  Authors: Antonio Carzaniga
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2003 University of Colorado
//  Copyright (C) 2005-2006 Antonio Carzaniga
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to one of the authors.
//
//
// $Id: maindoc.h,v 1.13 2010-03-11 15:16:44 carzanig Exp $
//
/** \mainpage Siena Fast Forwarding API Documentation

This documentation describes the application programmer interface of
the <a href="http://www.inf.usi.ch/carzaniga/siena/">Siena</a> Fast
Forwarding module.  Technical documentation including the design of
the forwarding algorithm and the general architecture of the
forwarding module is available in a <a
href="http://www.inf.usi.ch/carzaniga/cbn/forwarding/index.html#paper">technical
paper</a>.

<p>This implementation consists of a <em>fowarding table</em> and a
<em>driver</em> program.  A forwarding table is a one-to-one
association of <em>predicates</em> to <em>interfaces</em>.  A
predicate is a <em>disjunction</em> of <em>conjunctions</em> of
elementary <em>constraints</em>.  The forwarding table provides
essentially one match function that takes a <em>message</em>
<em>M</em> and outputs the set of interfaces associated with
predicates matched by <em>M</em>.  The data model of Siena includes
the definitions of \link siena::value values\endlink, \link
siena::attribute attributes\endlink, \link siena::message
messages\endlink, \link siena::constraint 
constraints\endlink, \link siena::filter filters\endlink, and \link
siena::predicate predicates\endlink, is presented in the \link types.h
types.h header file\endlink.

<p>The forwarding table can be used in two ways: 

<ol> 
<li>as an external program, through its <em>driver</em> program; and

<li>as an internal module, through its <em>application programming 
    interface</em>
</ol>

<p>The driver is a program that reads commands from an input file.
The commands understood by the driver include essentially all the
operations of the forwarding table.  In particular, the driver can add
associations of predicates to interfaces, and can request the
processing of messages.  An example including the main commands
accepted by the driver is available in the \link driver.input driver
input\endlink example from the <a href="examples.html">Example
Section</a>.  The syntax of the input to the driver is documented \ref
driversyntax "here".

<p>The API of the forwarding table is defined by the \link
siena::ForwardingTable ForwardingTable\endlink class.  This class is
an interface.  Seven concrete implementations are currently available:
\link siena::FwdTable FwdTable\endlink, \link siena::BTable
BTable\endlink, \link siena::SortedBTable SortedBTable\endlink, \link
siena::BXTable BXTable\endlink, \link siena::BCTable BCTable\endlink,
\link siena::BDDBTable BDDBTable\endlink, and \link siena::ZDDBTable
ZDDBTable\endlink.  These last six implementations (BTable,
SortedBTable, BXTable, BCTable, BDDBTable, and ZDDBTable) are based on
encoded predicates and messages.  Because of the nature of the
encoding, these implementations are likely to generate false
positives.  Therefore, applications using these implementations should
be designed to compensate for this behavior. (See \link siena::BTable
BTable\endlink for more details.)

In order to use the functionalities of the forwarding table,
applications must provide implementations of the Siena data model, by
implementing the siena::value, siena::attribute, siena::message,
siena::constraint, siena::filter, and siena::predicate interfaces.  A
complete sample implementation of these interfaces is available from
the <a href="examples.html">Example Section</a>.

**/

/** \example driver.input

Example of input commands for the driver of the forwarding table.
The syntax of commands is somehow similar to the syntax of C,
especially that of literal values.  

 **/

/** \page driversyntax Syntax of the Input to the Driver Program

This page contains a (almost complete) formal specification of the
syntax of the driver program.  An example is also available from
the <a href="examples.html">Example Section</a>:

<dl>
<dt>DriverInput<dd>
    Statement
    <br>DriverInput Statement

<dt>Statement<dd>
        <a href="#IFConfig">IFConfig</a>
    <br><a href="#Select">Select</a>
    <br><code>\%consolidate</code>
    <br><code>\%print_statistics</code>
    <br><code>\%statistics_only</code>
    <br><code>\%timer_start</code>
    <br><code>\%timer_stop</code>
    <br><code>\%output_on</code>
    <br><code>\%output_off</code>
    <br><code>\%output_to</code> <em>&nbsp;// to standard output</em>
    <br><code>\%output_to</code> <em>filename (C-style string)</em>
    <br><code>\%pre_process_limit</code>
    <br><code>\%print_statistics</code> <em>format (C-style string)</em>
    <br><code>\%clear</code>
    <br><code>\%clear_recycle</code>
    <br><code>\%pre_process_limit</code> <code>=</code> <a href="#Number">Number</a>

<dt>IFConfig<dd>
  <code>ifconfig</code> <a href="#Number">Number</a> Predicate

<dt>Predicate<dd>
    Constraint
    <br>Predicate Or Constraint
    <br>Predicate And Constraint

<dt>Or<dd> <code>||</code> <br><code>|</code> <br><code>\\/</code>

<dt>And<dd> <code>&&</code> <br><code>,</code> <br><code>/\\</code>

<dt>Constraint<dd> 
   <a href="#Name">Name</a> <a href="#Operator">Operator</a> <a href="#LiteralValue">LiteralValue</a>
   <br>Name <code>any</code> <a href="#Type">Type</a> <em>&nbsp;// any-value constraint</em>

<dt>Name<dd>
   {<code>A-Za-z</code>}{<code>A-Za-z/.-</code>}*

<dt>Operator<dd>
   <code>=</code> <br><code>&lt;</code> <br><code>&gt;</code>
   <br><code>=*</code> <em>&nbsp;// prefix</em>  
   <br><code>*=</code> <em>&nbsp;// suffix</em>
   <br><code>**</code> <em>&nbsp;// substring</em>

<dt>LiteralValue<dd>
   <em>C-style integer literal</em> 
   <br><em>C-style string literal</em> 
   <br><em>C-style double literal</em> 
   <br><code>true</code>
   <br><code>false</code>

<dt>Number<dd>
   <em>C-style integer literal</em> 

<dt>Type<dd> 
   <code>integer</code>
   <br><code>string</code>
   <br><code>boolean</code>
   <br><code>double</code>
   <br><code>any</code>

<dt>Select<dd>
  <code>select</code> Message

<dt>Message<dd>
   Attribute
   <br>Message Attribute

<dt>Attribute<dd>
   <a href="#Name">Name</a> <code>=</code> <a href="#LiteralValue">LiteralValue</a>

</dl>

**/

/** \example simple_types.cc

This is a complete implementation of the Siena data model.  In
particular, this file shows how to implement the the siena::value,
siena::attribute, siena::message, siena::constraint,
siena::filter, and siena::predicate interfaces.

**/

/** \example forwarding_messages.cc

This example shows how to set up a \link siena::MatchHandler match
handler\endlink for use with the \link siena::ForwardingTable
forwarding table\endlink.  In this example, the forwarding table
processes messages of type <code>text_message</code> delegating the
output function to an object of type <code>simple_handler</code>.
<code>simple_handler</code> puts out each message to a set of output
streams associated with the matching interfaces.  The handler also
implements a cut-off mechanism that limits the number of output
streams that a message is copied to.  In this particular example, this
limit is set to 5.  Notice that the handler is initialized with a
reference to message that is of the actual message type.  This allows
the handler to use specific access methods for its output function.
In this example, the handler uses the <code>get_text()</code> method
to obtain the output text from the message object.

**/

