// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.inf.usi.ch/carzaniga/siena/
//
//  Authors: Antonio Carzaniga
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2005 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to one of the authors.
//
//
// $Id: fwd_table.h,v 1.16 2010-03-11 15:16:40 carzanig Exp $
//
#ifndef FWD_TABLE_H
#define FWD_TABLE_H

#include <siena/exceptions.h>
#include <siena/forwarding.h>
#include <siena/allocator.h>

#include "bitvector.h"
#ifndef WITH_STATIC_COUNTERS
#include "patricia.h"
#endif
#include "bloom_set.h"

namespace siena {

class Filter;

/* link in a single-link list for filters */
class flist {
public:
    /* the filter */
    const Filter * f;
    /* link to the next filter in the list */
    const flist * next;

    /* constructor */
    flist(Filter * xf, const flist * xn) throw () : f(xf), next(xn) {};
};

/*  constraint descriptor
 *
 *  it is essentially a set of pointers to the filters in which this
 *  constraint appears.
 */
class Constraint {
public:
    /** @brief list of filters associated with this constraint
     *
     *  All the filters in which this constraint appears.
     *
     *  <p>Right now, we add filters to this list as they are passed
     *  to the forwarding table.  It might be possible and
     *  advantageous to figure out an optimal ordering of this list in
     *  order to speed up the maching process.  For example, we could
     *  sort the filters based on their size.
     **/
    const Filter * f;
    /** link to the next filter in the list **/
    const flist * next;

    /** @brief constructor **/
    Constraint() throw () : f(0), next(0) {};
};

/*  this constraint processor represents the main "iterator" of the
 *  matching function.
 */
class CProcessor {

public:
    /** initializes the constraint processor */
    CProcessor(unsigned int ifcount, MatchHandler & p, bitvector * v, 
	       const message & n) throw ()
	: if_mask(v), processor(p), target(ifcount) {
#ifdef WITH_STATIC_COUNTERS
	static unsigned int msg_id_factory = 0;
	msg_id = ++msg_id_factory;
#endif
	message::iterator * i  = n.first(); 
	if (i != 0) {
	    do {
		bset.add_string(i->name().begin, i->name().end);
	    } while (i->next());
	    delete(i);
	}
    }

    /** method called by the matching function for every matched
     *  constraint.
     *
     *  This method implements a fundamental part of the matching
     *  function.
     */
    bool		process_constraint(const Constraint *) throw ();

private:
    /** table of counters for partially-matched filters.
     *
     *  Each filter considered in the matching process has an
     *  associated counter that records the number of constraints
     *  matched so far for that filter.
     */
    // patricia_map<unsigned char> fmap;

    /** set of interfaces that can be ignored in the matching process.
     *
     *  An interface can be ignored by the matching process if it has
     *  already been matched and processed or if it was excluded by
     *  the pre-processing function.
     */
    bitvector *			if_mask;
    /** output processor
     */
    MatchHandler &		processor;
    /** total number of interfaces we can match.
     *
     *  we maintain this value so that we can stop the matching
     *  process immediately once we have matched all interfaces.
     */
    const unsigned int		target;

#ifdef WITH_STATIC_COUNTERS
    /** current message id
     */
    unsigned int		msg_id;
#else
    /** table of counters for partially-matched filters.
     *
     *  Each filter considered in the matching process has an
     *  associated counter that records the number of constraints
     *  matched so far for that filter.
     */
    patricia_map<unsigned char> fmap;
#endif
    /** Bloom set of the attribute names in the message being
     *  evaluated.
     *
     *  This Bloom set provides a "negative check" against filters.
     *  In essence, a message must contain all the attributes in a
     */
    bloom_set			bset;
};

} // end namespace siena

#endif
