// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.inf.usi.ch/carzaniga/siena/
//
//  Authors: Antonio Carzaniga
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2003 University of Colorado
//  Copyright (C) 2005 Antonio Carzaniga
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to one of the authors.
//
//
// $Id: bddbtable.h,v 1.12 2010-03-11 15:16:44 carzanig Exp $
//
#ifndef SIENA_BDDBTABLE_H
#define SIENA_BDDBTABLE_H

#include <cstddef>

#include <siena/forwarding.h>
#include <siena/allocator.h>

namespace siena {

/** @brief implementation of the forwarding table based on Bloom
 *         filters and BDDs. This is the common structure of the
 *         BDD-based algorithm.
 *
 *  The idea behind this algorithm is to represent each encoded
 *  predicate as a BDD.  An encoded predicate is constructed as an
 *  M-by-N bit matrix, where M is the chosen size of the Bloom filter,
 *  and N is the number of Bloom filters, each one representing a
 *  conjunction of constraints.  Therefore, each one of the M bits of
 *  the Bloom filters becomes a variable in the BDD (or ZDD).
 **/
class BDDBTableBase : public ForwardingTable {
public:
    virtual void consolidate() throw();

    virtual void ifconfig(if_t, const predicate &) throw(bad_constraint) = 0;

    virtual void match(const message &, MatchHandler &) const throw() = 0;
    virtual void match(const message & m, MatchMessageHandler & p) const throw();
    virtual size_t allocated_bytesize() const throw();
    virtual size_t bytesize() const throw();

    virtual void clear() throw() = 0;
    virtual void clear_recycle() throw() = 0;

protected:
    /** @brief Protected allocator of the forwarding table.  
     *
     *  All the data structure forming the forwarding table are
     *  allocated through this memory management system.
     **/
    FTAllocator memory;
};

/** @brief implementation of the forwarding table based on Bloom
 *         filters. This implementation consolidates each predicate
 *         into a BDD, which is then used for matching.
 *
 *  @see BDDBTableBase
 **/
class BDDBTable : public BDDBTableBase {
public:
    BDDBTable() throw();
    virtual ~BDDBTable() throw();

    virtual void ifconfig(if_t, const predicate &) throw(bad_constraint);
    virtual void match(const message &, MatchHandler &) const throw();

    virtual void clear() throw();
    virtual void clear_recycle() throw();

private:
    class IFDescr;

    IFDescr * bddlist;
};

/** @brief implementation of the forwarding table based on Bloom
 *         filters. This implementation consolidates each predicate
 *         into a ZDD, which is then used for matching.
 *
 *  @see BDDBTableBase
 **/
class ZDDBTable : public BDDBTableBase {
public:
    ZDDBTable() throw();
    virtual ~ZDDBTable() throw();

    virtual void ifconfig(if_t, const predicate &) throw(bad_constraint);
    virtual void match(const message &, MatchHandler &) const throw();

    virtual void clear() throw();
    virtual void clear_recycle() throw();

private:
    class IFDescr;

    IFDescr * zddlist;
};

}; // end namespace siena

#endif
