// -*- C++ -*-
//
//  This file is part of Siena, a wide-area event notification system.
//  See http://www.inf.usi.ch/carzaniga/siena/
//
//  Authors: Antonio Carzaniga
//  See the file AUTHORS for full details. 
//
//  Copyright (C) 2001-2003 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to one of the authors.
//
//
// $Id: bxtable.h,v 1.6 2010-03-11 15:16:44 carzanig Exp $
//
#ifndef SIENA_BXTABLE_H
#define SIENA_BXTABLE_H

#include <cstddef>

#include <siena/forwarding.h>
#include <siena/allocator.h>

namespace siena {

class BXPredicate;

/** @brief implementation of the forwarding table based on Bloom
 *         filters and a data structure used for matching called XDD.
 *
 *  Like BTable, this implementation is based on encoded filters and
 *  messages.  However, instead of using a linear scan of the set of
 *  encoded filters, this implementation indexes the bitvectors
 *  representing filters with a decision diagram. 
 *
 *  \b WARNING: This implementation is provided primarily for
 *  experimental purposes.  See BTable for details.
 **/
class BXTable : public ForwardingTable {
public:
    BXTable() throw();

    virtual ~BXTable() throw();

    virtual void ifconfig(if_t, const predicate &) throw(bad_constraint);
    virtual void consolidate() throw();

    virtual void match(const message &, MatchHandler &) const throw();
    virtual void match(const message &, MatchMessageHandler &) const throw();

    virtual void clear() throw();
    virtual void clear_recycle() throw();

    virtual size_t allocated_bytesize() const throw();
    virtual size_t bytesize() const throw();

protected:
    /** @brief Protected allocator of the forwarding table.  
     *
     *  All the data structure forming the forwarding table are
     *  allocated through this memory management system.
     **/
    FTAllocator		memory;

    /** @brief list of predicates.
     *
     *  each link in this linked list is a pair <predicate,interface-id>
     **/
    BXPredicate *	plist;
};

} // end namespace siena

#endif
